"""
    ZCGI
    ====
"""
# $Id: ZCGI.py,v 1.7 2002/01/04 23:38:03 adrian Exp $
#
# $Log: ZCGI.py,v $
# Revision 1.7  2002/01/04 23:38:03  adrian
# Fixed CGI on Windows "COMSPEC" bug
#
# Revision 1.6  2001/12/31 22:19:26  adrian
# Fixed a few minor bugs, added online documentation and a README
#
# Revision 1.5  2001/12/31 20:56:26  adrian
# Fixed bug in ZCGIScript triggered by incorrect cgiscript name or missing
# zope cgi-bin folder.
#
# Revision 1.4  2001/12/30 01:05:44  adrian
# * Added ZCGIFolder type to emulate CGI-BIN functionality exactly
# * Corrected environment for CGI script
# * Corrected passing POST data
# * Corrected handling of returned mime-headers
# * Added HTML-HEAD ripping (Can change CGI data to fit site look&feel)
# * Added code to fix mailman malformed HTML (spurious <BODY in output)
# * Changed error messages to be more standard for CGI-BIN
# * Can now add new ZCGIExtensionMap's to allow certain handlers only in certain areas
#
# Revision 1.3  2001/12/11 08:05:19  adrian
# Partially working
#
# Revision 1.2  2001/10/21 15:04:56  adrian
# no message
#
#
__version__ = '$Revision: 1.7 $'[11:-2]

import Globals
import OFS.SimpleItem
import OFS.PropertyManager
import AccessControl.Role
from DocumentTemplate import DT_HTML
from App.Extensions import getObject, getPath, FuncCode
from DocumentTemplate import DT_HTML
import copy
import os
import sys
import string
import urllib
import Zope
import re

#####################
# Regular Expressions
#
# Used for ripping HTML pages apart
#HTML = re.compile('<BODY.*?>(.*)</BODY>', re.M)
GET_HEAD = re.compile('<head.*?>(.*)</head>', re.I | re.M | re.S)
GET_SCRIPTS = re.compile('</head>(.*)<body.*?>', re.I | re.M | re.S)
GET_BODY = re.compile('<body.*?>(.*)(</body>)?', re.I | re.M | re.S)
GET_TITLE = re.compile('<title.*?>(.*)</title>', re.I | re.M | re.S)
MAILMAN_BUG = re.compile('<body.*?(<body.*?>)', re.I | re.M | re.S)

################
# CGI-BIN Script
#

manage_addZCGIScriptForm = Globals.HTMLFile('www/addZCGIScript', globals())
def manage_addZCGIScript(self, id, title='', cgifile='', submit='', REQUEST=None):
    """ Add a ZCGIScript object """
    obj = ZCGIScript(id, title=str(title), cgifile=str(cgifile))
    self._setObject(obj.getId(), obj)
    if REQUEST is not None:
        try: u=self.DestinationURL()
        except: u=REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u, urllib.quote(obj.getId()))
        REQUEST.RESPONSE.redirect(u+'/manage_workspace')
    return ''

class ZCGIScript(
        OFS.PropertyManager.PropertyManager,
        AccessControl.Role.RoleManager,
        OFS.SimpleItem.SimpleItem,
        ):
    """ A product to allow Zope to a single cgi-bin script """
    meta_type = 'ZCGIScript'
    defaulttitle = 'Zope CGI Script'
    cgifile = ''
    headers = 'CGI'
    manage_options    = (
        OFS.PropertyManager.PropertyManager.manage_options
        +(
            {'label': 'View', 'action': 'preview',
             'help': ('ZCGI', 'preview.stx')},
            {'label': 'Run',  'action': ''},
        )
        +OFS.SimpleItem.SimpleItem.manage_options
        +AccessControl.Role.RoleManager.manage_options
    )
    __ac_permissions__ = (
        (
            ('View',
                ('',),
                ('Anonymous', 'Manager',),
            ),
            ('Manage CGI-BIN Scripts',
                ('preview',),
                ('Manager',),
            ),
        )
    )
    _properties=(
        {'id' : 'defaulttitle', 'type': 'string', 'mode': 'w'},
        {'id' : 'cgifile', 'type' : 'string', 'mode': 'w'},
        {'id' : 'headers', 'type' : 'selection', 'mode' : 'w', 'select_variable' : 'headerType'},
    )

    def headerType(self):
        return (
            'CGI',
            'Zope',
            'Manage',
        )

    def __init__(self, id, title='', cgifile=''):
        """ Initialize a new ZCGIScript object """
        self.id = id
        self.defaulttitle = title
        self.cgifile = cgifile

    def title_or_id(self):
        """ """
        if self.title():
            return self.title()
        else:
            return self.getId()

    def title_and_id(self):
        """ """
        if self.title():
            return "%s (%s)" % (self.title(), self.getId())
        else:
            return "%s" % (self.getId(),)

    def title(self):
        """ Return a title for this object """
        if self.REQUEST.has_key('title'):
            return self.REQUEST['title']
        else:
            return self.defaulttitle

    def __repr__(self, overrideZopeHeaders=0, overrideManageHeaders=0):
        """ run CGI-BIN program. """
        # Find the file and handler
        name, handler = self._getFileAndHandler()
        # Backup the environment - We are about to break it!
        tmpenv = copy.deepcopy(os.environ)
        # Sanitize the environment for the script
        for k in os.environ.keys():
            if string.lower(k) not in ('path', 'comspec', 'temp', 'tmp'):
                os.environ[k] = ''
                del os.environ[k]
        for k in self.REQUEST.environ.keys():
            if '.' not in k:
                os.environ[k] = self.REQUEST.environ[k]
        tmp = string.replace(self.REQUEST['URL0'], self.REQUEST['BASE0'], '')
        if self.REQUEST['QUERY_STRING']:
            tmp = tmp + "?" + str(self.REQUEST['QUERY_STRING'])
        os.environ['REQUEST_URI'] = tmp
        os.environ['PATH_TRANSLATED'] = os.path.dirname(name) + os.environ['PATH_INFO']
        os.environ['DOCUMENT_ROOT'] = os.path.dirname(name)
        os.environ['SERVER_ADDR'] = string.split(os.environ['HTTP_HOST'], ':')[0]
        os.environ['SCRIPT_FILENAME'] = name
        data = urllib.urlencode(self.REQUEST.form)
        try:
            w,r,e=os.popen3(handler % name)
            w.write(data)
            w.flush()
            w.close()
            retLines = []
            for line in r.readlines():
                if line[-1] in ("\r", "\n"):
                    line = line[:-1]
                retLines.append(line)
            errs = e.read()
        except:
            retLines = ['content-type: text/plain', '', 'Error!!']
            errs = "An error occured"
#        retLines = ['content-type: text/html', '']
#        for k in os.environ.keys():
#            retLines.append('%s = %s<br>' % (k, os.environ[k]))
#        errs=''
        # Repair the environment
        for k in os.environ.keys():
            del os.environ[k]
        for k in tmpenv.keys():
            os.environ[k] = tmpenv[k]
        # If it returned data via STDERR then format a nice error message
        if errs:
            raise "Internal Server Error", "Script returned error:<br><pre>%s</pre>" % (errs, )
        # Get the headers from the returned data, raise "Internal Server Error" if there isn't at least "content-type"
        try:
            hLines = []
            while retLines[0]:
                hLines.append(retLines.pop(0))
            retLines.pop(0)
            for line in hLines:
                k, v = string.split(line, ": ", 1)
                self.REQUEST.RESPONSE.setHeader(k, v)
        except:
            raise "Internal Server Error", "Error decoding script headers"
        # If headers is set to "Zope" or "Manage", replace the head section and
        # bottom with either standard_html_header/footer or manage_page_header/footer
        # Rip head content out, and place it after the new header (Incorrect, but it works)
        # Only do this for text/html of course.
        html = string.join(retLines, "\n")
        if self.REQUEST.RESPONSE.getHeader('content-type') == 'text/html':
            html = MAILMAN_BUG.sub('\g<1>', html)
            if self.headers != 'CGI':
                h = GET_HEAD.search(html)
                if h:
                    h = string.join(h.groups())
                    t = GET_TITLE.search(h)
                    if t:
                        t = string.join(t.groups())
                    else:
                        t = self.title
                    h = GET_TITLE.sub('', h)
                else:
                    h = ''
                    t = self.defaulttitle
                self.REQUEST.set('title', t)
                s = GET_SCRIPTS.search(html)
                if s:
                    s = string.join(s.groups())
                else:
                    s = ''
                b = GET_BODY.search(html)
                if b:
                    b = b.group(1)
                else:
                    return html
                if self.headers == 'Zope':
                    html = getattr(self, 'standard_html_header')(self, REQUEST=self.REQUEST) + h + s + b + getattr(self, 'standard_html_footer')(self, REQUEST=self.REQUEST)
                else:
                    html = getattr(self, 'manage_page_header')(self, REQUEST=self.REQUEST) + h + s + b + getattr(self, 'manage_page_footer')(self, REQUEST=self.REQUEST)
        return html

    def _getFileAndHandler(self):
        """ Find the cgi-bin file """
        if os.path.split(self.cgifile)[0]:
            raise ValueError, ('The file name, %s, should be a simple file name' % self.cgifile)
        if os.path.splitext(self.cgifile)[1]:
            raise ValueError, ('The file name, %s, should not include an extension' % self.cgifile)
        extmap = getattr(self, ZCGIExtensionMap.id)
        for home in (INSTANCE_HOME, os.path.split(os.path.split(SOFTWARE_HOME)[0])[0]):
            rootfilename = os.path.join(os.path.join(home, 'cgi-bin'), self.cgifile)
            for ext in extmap.getExtMap():
                if ext[0]:
                    file="%s.%s" % (rootfilename, ext[0])
                else:
                    file = rootfilename
                if os.path.exists(file):
                    return (file, ext[1])
        raise ValueError, ('There is no cgi-bin program with the rootname %s with an extension matching those in the ZCGIExtensionMap' % self.cgifile)

    def _bt_processing(self, TraversalRequest, name):
        """ Called by __bobo_traverse__ (In the 'Magic' section) - ZCGIScript Version """
        # If there's no more names left to handle, return __repr__
        # to the traversal machinery so it gets called next
        if len(TraversalRequest['TraversalRequestNameStack'])==0:
            TraversalRequest.environ['PATH_INFO'] = string.join([''] + TraversalRequest['PATH_STACK'], '/')
            TraversalRequest.environ['SCRIPT_NAME'] = string.replace(self.absolute_url(), TraversalRequest['BASE0'], '')
            TraversalRequest.environ['REQUEST_URI'] = string.replace(TraversalRequest['URL0'], TraversalRequest['BASE0'], '')
            return getattr(self, '__repr__')
        # If there are more names to handle, return ourself, so our __bobo_traverse__ gets
        # called again and we get to ignore the other names.
        return self

    def __bobo_traverse__(self, TraversalRequest, name):
        """ Used in handling PATH_INFO, that is filepath information after the script name """
        if not TraversalRequest.has_key('PATH_STACK'):
            TraversalRequest['PATH_STACK'] = []
            TraversalRequest['REMOVE_STACK'] = []
        # If name is a method/attribute of this object, call it
        TraversalRequest['REMOVE_STACK'].append(name)
        if name!='index_html':
            TraversalRequest['PATH_STACK'].append(name)
            if hasattr(self.aq_base, name):
                return getattr(self, name)
        # otherwise do some 'magic'
        return self._bt_processing(TraversalRequest, name)

    def preview(self):
        """ Return the output from the cgi-bin and process it as DTML """
        cgifile = self._getFileAndHandler()[0]
        dtml = "<dtml-var manage_page_header><dtml-var manage_tabs><h1>Testing cgi-bin</h1>"
        dtml = dtml + "<h2>Running %s</h2><hr>" % (cgifile, )
        dtml = dtml + repr(self)
        dtml = dtml + "<hr><dtml-var manage_page_footer>"
        obj = DT_HTML.HTML(dtml, self.REQUEST, 'ZCGI - %s - %s' % (cgifile, self.getId(),))
        return obj(self.aq_inner, self.REQUEST)

################
# CGI-BIN Folder
#

manage_addZCGIFolderForm = Globals.HTMLFile('www/addZCGIFolder', globals())
def manage_addZCGIFolder(self, id, title='', cgifolder='', submit='', REQUEST=None):
    """ Add a ZCGIFolder object """
    obj = ZCGIFolder(id, title=str(title), cgifolder=str(cgifolder))
    self._setObject(obj.getId(), obj)
    if REQUEST is not None:
        try: u=self.DestinationURL()
        except: u=REQUEST['URL1']
        if submit==" Add and Edit ": u="%s/%s" % (u, urllib.quote(obj.getId()))
        REQUEST.RESPONSE.redirect(u+'/manage_workspace')
    return ''

class ZCGIFolder(
        ZCGIScript,
        ):
    """ A product to allow Zope to call cgi-bin executables """
    meta_type = 'ZCGIFolder'
    defaulttitle = 'Zope CGI Folder'
    cgifolder = ''
    manage_options    = (
        OFS.PropertyManager.PropertyManager.manage_options
        +OFS.SimpleItem.SimpleItem.manage_options
        +AccessControl.Role.RoleManager.manage_options
    )
    __ac_permissions__ = (
        (
            ('View',
                ('',),
                ('Anonymous', 'Manager',),
            ),
            ('Manage CGI-BIN Scripts',
                ('preview',),
                ('Manager',),
            ),
        )
    )
    _properties=(
        {'id' : 'defaulttitle', 'type': 'string', 'mode' : 'w'},
        {'id' : 'cgifolder', 'type' : 'string', 'mode' : 'w'},
        {'id' : 'headers', 'type' : 'selection', 'mode' : 'w', 'select_variable' : 'headerType'},
    )

    def __init__(self, id, title='', cgifolder=''):
        """ Initialize a new ZCGIFolder object """
        self.id = id
        self.defaulttitle = title
        self.cgifolder = cgifolder

    def _getFileAndHandler(self):
        """ Find the cgi-bin file """
        basename = string.split(self.REQUEST['SCRIPT_NAME'], '/')[-1]
        extmap = getattr(self, ZCGIExtensionMap.id)
        if os.path.splitext(basename)[1]:
            raise 'NotFound'#, ('The file name, %s, should not include an extension' % basename)
        rootfilename = os.path.join(self.cgifolder, basename)
        for ext in extmap.getExtMap():
            if ext[0]:
                file="%s.%s" % (rootfilename, ext[0])
            else:
                file = rootfilename
            if os.path.exists(file):
                return (file, ext[1])
        raise 'NotFound'#, ('There is no cgi-bin program with the rootname %s with an extension matching those in the ZCGIExtensionMap' % basename)

    def _bt_processing(self, TraversalRequest, name):
        """ Called by __bobo_traverse__ (In the 'Magic' section) - ZCGIFolder Version """
        # If there's no more names left to handle, return __repr__
        # to the traversal machinery so it gets called next
        if  len(TraversalRequest['TraversalRequestNameStack'])==0:
            if not TraversalRequest.has_key('script'):
                TraversalRequest['script'] = TraversalRequest['PATH_STACK'].pop(0)
            TraversalRequest.environ['PATH_INFO'] = string.join([''] + TraversalRequest['PATH_STACK'], os.sep)
            TraversalRequest.environ['SCRIPT_NAME'] = string.replace(self.absolute_url() + '/' + TraversalRequest['script'], TraversalRequest['BASE0'], '')
            return getattr(self, '__repr__')
        # If there are more names to handle, return ourself, so this __bobo_traverse__ gets
        # called again and we get to ignore the other names.
        return self

################################
# Mapping extensions to handlers
#

def manage_addZCGIExtensionMap(self, REQUEST=None):
    """ Add a ZCGIExtensionMap object """
    obj = ZCGIExtensionMap()
    self._setObject(obj.getId(), obj)
    if REQUEST is not None:
        u=self.DestinationURL()
        REQUEST.RESPONSE.redirect(u+'/manage_workspace')
    return ''

class ZCGIExtensionMap(
        AccessControl.Role.RoleManager,
        OFS.SimpleItem.SimpleItem,
        ):
    """ An extension map for ZCGI to aquire """
    id          =   "ZCGIExtMap"
    title       =   "Z CGI Extension Map"
    manage_options    = (
        (
            {'label': 'Extension Map', 'action': 'manage_extensionMapForm',
             'help': ('ZCGI', 'extensionMap.stx')},
        )
        +OFS.SimpleItem.SimpleItem.manage_options
        +AccessControl.Role.RoleManager.manage_options
    )
    manage_extensionMapForm = Globals.HTMLFile('www/extensionMap', globals())

    def __init__(self):
        """ Initialise with a default extension map """
        self.extmap = default_extension_map

    def manage_saveExtMap(self, REQUEST=None):
        """ Reorder/Delete/Save Changes in the top part of the form """
        exts1 = []
        exts2 = []
        getreq = self.REQUEST.get
        submit = getreq('manage_saveExtMap', 'None Specified')
        for ext in self.extmap:
            if submit == "Move First":
                if getreq('f_%s' % ext[0], ''):
                    exts1.append(ext)
                else:
                    exts2.append(ext)
            elif submit == "Delete":
                if not getreq('f_%s' % ext[0], ''):
                    exts1.append(ext)
            elif submit == "Save Changes":
                exts1.append(
                    (ext[0], getreq('h_%s' % ext[0], ext[1])),
                )
            else:
                raise "Parameter error", "Don't know how to '%s' on this form" % submit
            self.extmap = tuple(exts1 + exts2)
        if REQUEST:
            message="Z CGI Extension Map Updated."
            return self.manage_extensionMapForm(self, REQUEST, manage_tabs_message=message)

    def manage_addExt(self, ext='', handler='', REQUEST=None):
        """ """
        extentry = (ext, handler)
        exts = map(None, self.extmap)
        exts.append(extentry)
        self.extmap = tuple(exts)
        if REQUEST:
            message="Handler Added."
            return self.manage_extensionMapForm(self, REQUEST, manage_tabs_message=message)

    def getExtMap(self):
        """ Return the extension map """
        return self.extmap

if sys.platform=='win32':
    default_extension_map = (
        ('exe', '%s'),
        ('com', '%s'),
        ('bat', 'command.com /c %s'),
        ('cmd', 'cmd.exe /c %s'),
        ('pl', 'perl.exe %s'),
        ('py', 'python.exe %s'),
    )
else:
    default_extension_map = (
        ('sh', 'sh %s'),
        ('pl', 'perl %s'),
        ('py', 'python %s'),
        ('', '%s'),
    )

if not hasattr(os, 'popen3'):
    raise "Module os does not contain popen3. Can not continue"
